/*     
 * $RCSfile: semaphore.c,v $
 *
 * x-kernel v3.3
 *
 * Copyright (c) 1993,1991,1990,1996  Arizona Board of Regents
 *
 * $Log: semaphore.c,v $
 * Revision 1.6  1997/03/17 06:36:51  dorgival
 * Changed calls to lock/unlock the x-kernel to fix deadlock condition
 * Removed extra while in semWait which should not be thre
 *
 * Revision 1.5  1997/03/17  01:32:11  dorgival
 * Added macros to replace pthread_mutex/cond calls
 *
 * Revision 1.4  1997/03/16  18:28:10  dorgival
 * Increased tracing info with this_thread->status
 * Started breaking xk_lock to make mutex locking external (USE_SEM_MUTEX)
 *
 * Revision 1.3  1997/03/12 22:47:04  dorgival
 * Just replaced xk_all_unlock with xk_any_unlock
 *
 * Revision 1.2  1997/03/11 20:57:59  dorgival
 * Created xk_thread_t and updated threadSelf()
 *
 * Revision 1.1  1997/03/11 20:47:08  dorgival
 * Initial revision
 *
 *
 */

/*
 * Counting semaphores implemented on top of threads package.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <assert.h>
#include "xk_semaphore.h"
#include "xk_thread.h"
#include "trace.h"
#include "xk_debug.h"
#include "event.h"
#include "x_util.h"

int tracesemaphores;

void semDestroy( Semaphore* s );

static void semCheckSignature( Semaphore* s );

static void
semCheckSignature( Semaphore* s )
{
    if ( s->sem_signature != s ) {
	xTrace1(semaphores,TR_ALWAYS,"semCheckSignature: sem 0x%x", s);
    }
#ifdef USE_SEM_MUTEX
    if ( s->mutex_signature != &s->mutex ) {
	xTrace1(semaphores,TR_ALWAYS,"semCheckSignature: sem 0x%x mutex", s );
    }
#endif
    if ( s->cond_signature != &s->cond ) {
	xTrace1(semaphores,TR_ALWAYS,"semCheckSignature: sem 0x%x cond", s );
    }
}

/*
 * wake_sem is used to implement Delay... 
 */
static void wake_sem( Event ev, VOID *arg );

void
semInit(Semaphore* s, u_int initialVal)
{
    xTrace3(semaphores,TR_FUNCTIONAL_TRACE,"semInit(0x%x,%d) by thread 0x%x",
            s,initialVal,threadSelf() );
    s->sem_signature   = s;
#ifdef USE_SEM_MUTEX
    s->mutex_signature = &s->mutex;
#endif
    s->cond_signature  = &s->cond;
    s->count = initialVal;
#ifdef USE_SEM_MUTEX
    pthread_mutex_init(&(s->mutex), pthread_mutexattr_default);
#endif
    pthread_cond_init(&(s->cond), pthread_condattr_default);
    semCheckSignature(s);
}

void
semDestroy( Semaphore* s )
{
    xTrace2(semaphores,TR_FUNCTIONAL_TRACE,"semDestroy(0x%x) by thread 0x%x",
            s,threadSelf() );
    semCheckSignature(s);
#ifdef USE_SEM_MUTEX
    if ( pthread_mutex_destroy(&(s->mutex)) < 0 ) {
	if ( errno == EBUSY ) {
	    xTrace2(semaphores,TR_SOFT_ERRORS,
	            "semDestroy(0x%x) failed (thread 0x%x): mutex is locked",
		    s,threadSelf() );
        } else {
	    xTrace2(semaphores,TR_SOFT_ERRORS,
	            "semDestroy(0x%x) failed (thread 0x%x): invalid mutex?",
		    s,threadSelf() );
	}
    }
#endif
    if ( pthread_cond_destroy(&(s->cond)) < 0 ) {
	if ( errno == EBUSY ) {
	    xTrace2(semaphores,TR_SOFT_ERRORS,
	    "semDestroy(0x%x) failed (thread 0x%x): cond is being waited on",
		    s,threadSelf() );
        } else {
	    xTrace2(semaphores,TR_SOFT_ERRORS,
	            "semDestroy(0x%x) failed (thread 0x%x): invalid cond?",
		    s,threadSelf() );
	}
    }
}

/*
 * Wait and signal are simplified by the way condition variables are
 * handled by the p-threads package: the wait/signal calls already
 * deal with the problem of releasing/acquiring a mutex lock.
 * All that is necessary is to pass the master mutex in those calls.
 */

void
semWait(Semaphore* s)
{
    xk_thread_t* this_thread = threadSelf();

    xTrace2(semaphores,TR_FUNCTIONAL_TRACE,"semWait(0x%x) by thread 0x%x",
            s,this_thread );
    semCheckSignature(s);
    assert(xk_lock_owner() == this_thread);

#ifdef USE_SEM_MUTEX
    PTHREAD_MUTEX_LOCK(&(s->mutex));
#endif

    if (--s->count < 0) {
	xTrace2(semaphores,TR_FUNCTIONAL_TRACE,
	        "semWait(0x%x) by thread 0x%x will really wait",
		s,this_thread );
	SET_THREAD_STATUS( this_thread, XK_SEM_WAIT );
#ifdef USE_SEM_MUTEX
	XK_UNLOCK();

	PTHREAD_COND_WAIT(&(s->cond),&(s->mutex));

	xTrace2(semaphores,TR_FUNCTIONAL_TRACE,
	        "semWait(0x%x) by thread 0x%x back from wait",s,this_thread);
	XK_LOCK( XK_INTERNAL_LOCK );
#else
	PTHREAD_MUTEX_LOCK(xk_mutex);
	xk_any_unlock();
	PTHREAD_COND_WAIT(&(s->cond),xk_mutex);
	xk_any_lock( XK_INTERNAL_LOCK );
	PTHREAD_MUTEX_UNLOCK(xk_mutex);
#endif
	RESET_THREAD_STATUS( this_thread, XK_SEM_WAIT );
    }
    SET_THREAD_STATUS( this_thread, XK_SEM_OWN );
#ifdef USE_SEM_MUTEX
    PTHREAD_MUTEX_UNLOCK(&(s->mutex));
#endif
    xTrace2(semaphores,TR_FUNCTIONAL_TRACE,
            "semWait(0x%x): thread 0x%x returning", s,this_thread );
} 

void
semSignal(Semaphore* s)
{
    xk_thread_t* this_thread = threadSelf();

    xTrace2(semaphores,TR_FUNCTIONAL_TRACE,"semSignal(0x%x) by thread 0x%x",
            s,this_thread );
    semCheckSignature(s);
    assert(xk_lock_owner() == this_thread);

#ifdef USE_SEM_MUTEX
    PTHREAD_MUTEX_LOCK(&(s->mutex));
#endif

    RESET_THREAD_STATUS( this_thread, XK_SEM_OWN );
    if (s->count++ < 0) {
	xTrace2(semaphores,TR_FUNCTIONAL_TRACE,
	        "semSignal(0x%x) by thread 0x%x signalling other threads",
		s,this_thread );
	pthread_cond_signal(&(s->cond));
    }
#ifdef USE_SEM_MUTEX
    PTHREAD_MUTEX_UNLOCK(&(s->mutex));
#endif
    xTrace2(semaphores,TR_FUNCTIONAL_TRACE,
            "semSignal(0x%x) thread 0x%x returning", s,this_thread );
}


int
semCount (Semaphore* s)
{
    semCheckSignature(s);
    return s->count;
}


extern void
semSignalAll (Semaphore* s)
{
    int i, nblocked = -s->count;

    xTrace3(semaphores,TR_FUNCTIONAL_TRACE,
            "semSignalAll(0x%x) by thread 0x%x will wake %d threads",
	    s,threadSelf(), nblocked );
    semCheckSignature(s);
    for (i = 0; i < nblocked; ++i) {
	semSignal(s);
    }
    xTrace2(semaphores,TR_FUNCTIONAL_TRACE,
            "semSignalAll(0x%x) by thread 0x%x returning",s,threadSelf() );
}

void
Delay( int n )
{
    Semaphore s;
    xk_thread_t* this_thread = threadSelf();

    xTrace2(semaphores,TR_MAJOR_EVENTS,
            "Delay(%d) by thread 0x%x",n,this_thread );
    assert(xk_lock_owner() == this_thread );
    SET_THREAD_STATUS( this_thread, XK_DELAYED );
    semInit(&s, 0);
    evDetach( evSchedule(wake_sem, &s, n*1000) );
    semWait(&s);
    RESET_THREAD_STATUS( this_thread, XK_DELAYED );
    semDestroy(&s);
    assert(xk_lock_owner() == threadSelf());
}

static void
wake_sem( Event ev, VOID *arg )
{
  semSignal((Semaphore *)arg);
}
